import functools
import sys
import threading
import typing as tp

from satella.coding.structures import Singleton
from satella.instrumentation import Traceback
from .exception_handlers import BaseExceptionHandler, FunctionExceptionHandler, \
    exception_handler

__all__ = [
    'GlobalExcepthook'
]


@Singleton
class GlobalExcepthook:
    """
    Global event interceptor.

    Always installed
    """

    def __init__(self):
        self.installed_hooks = []  # list of BaseExceptionHandler
        self.ignore_failed_hooks = True
        self.__install()

    def remove_hook(self, hook: BaseExceptionHandler):
        """
        Unregister a hook

        :param hook: hook to remove
        :raise ValueError: if hook not in list
        """
        self.installed_hooks.remove(hook)

    def add_hook(self, new_hook: tp.Union[
        tp.Callable, BaseExceptionHandler]) -> BaseExceptionHandler:
        """
        Register a hook to fire in case of an exception.

        :param new_hook: callable(type, value, traceback) or instance of BaseExceptionHandler to use
        """

        if not isinstance(new_hook, BaseExceptionHandler):
            new_hook = FunctionExceptionHandler(new_hook)

        if new_hook not in self.installed_hooks:
            self.installed_hooks.append(new_hook)

        return new_hook

    def __install(self):
        """Enable _GlobalExceptHook to intercept exception_handling"""

        self.old_excepthook = exception_handler(priority=10000)(sys.excepthook)

        sys.excepthook = self.__except_handle

        # see http://bugs.python.org/issue1230540 for explaination to below
        init_old = threading.Thread.__init__

        my_self = self

        @functools.wraps(threading.Thread.__init__)
        def init(self, *args, **kwargs):  # custom wrapper for Thread's.__init__
            init_old(self, *args, **kwargs)
            run_old = self.run  # we will need to wrap thread's run() to catch exception_handling..

            @functools.wraps(run_old)
            def run_with_except_hook(*args, **kw):  # our new run
                # noinspection PyBroadException
                try:
                    return run_old(*args, **kw)
                except Exception as e:
                    try:
                        my_self.__except_handle(
                            *sys.exc_info())  # by now, it's our handler :D
                    except AttributeError:
                        pass  # Python interpreter is in an advanced state of shutdown, just let it go

                    raise  # re-raise if running on debug

                    # other BaseException's will be raised

            self.run = run_with_except_hook

        threading.Thread.__init__ = init

    def __except_handle(self, type_, value, traceback) -> None:
        hooks_to_run = self.installed_hooks + [self.old_excepthook]

        for hook in sorted(hooks_to_run, key=lambda h: h.priority):
            # noinspection PyBroadException
            try:
                if hook.handle_exception(type_, value, traceback):
                    break
            except Exception as e:
                tb = Traceback()
                sys.stderr.write('Error while processing a hook: \n')
                tb.pretty_print(output=sys.stderr)

                if not self.ignore_failed_hooks:
                    raise NotImplementedError(
                        'Hook failed (%s), but ignore_failed_hooks was false' % (
                            repr(e)))


GlobalExcepthook()



from .dict_object import *
from .immutable import Immutable, frozendict
from .singleton import Singleton, SingletonWithRegardsTo
from .structures import *
from .typednamedtuple import *

__all__ = [
    'typednamedtuple',
    'Heap',
    'SetHeap',
    'frozendict'
    'TimeBasedHeap',
    'OmniHashableMixin',
    'Singleton',
    'SingletonWithRegardsTo',
    'DictObject',
    'apply_dict_object',
    'Immutable'
]



import logging
import typing as tp
import warnings

logger = logging.getLogger(__name__)


def infinite_counter(start_at: int = 0, step: int = 1) -> tp.Generator[int, None, None]:
    """
    Infinite counter, starting at start_at

    :param start_at: value at which to start counting. It will be yielded as first
    :param step: step by which to progress the counter
    """
    warnings.warn('This is deprecated, use itertools.count() instead', DeprecationWarning)
    i = start_at
    while True:
        yield i
        i += step


def is_instance(classes: tp.Union[tp.Tuple[type, ...], type]) -> tp.Callable[[object], bool]:
    def inner(object_):
        return isinstance(object_, classes)
    inner.__doc__ = """Return a bool telling if object is of type %s""" % (repr(classes), )
    return inner


T = tp.TypeVar('T')


def take_n(iterator: tp.Iterable[T], n: int, skip: int = 0) -> tp.List[T]:
    """
    Take (first) n elements of an iterator, or the entire iterator, whichever comes first

    :param iterator: iterator to take from
    :param n: amount of elements to take
    :param skip: elements from the start to skip
    :return: list of length n (or shorter)
    """
    iterator = iter(iterator)
    for i in range(skip):
        next(iterator)

    output = []
    for i in range(n):
        try:
            output.append(next(iterator))
        except StopIteration:
            return output
    return output



import ctypes
import logging
import platform
import threading

logger = logging.getLogger(__name__)


class TerminableThread(threading.Thread):
    """
    Class that will execute something in a loop unless terminated. Use like:

    >>> class MeGrimlock(TerminableThread):
    >>>     def loop(self):
    >>>         ... do your operations ..
    >>> a = MeGrimlock()
    >>> a.start()
    >>> a.terminate().join()

    Flag whether to terminate is stored in **self._terminating**.

    If you decide to override run(), you got to check periodically for **self._terminating** to
    become true.
    If you decide to use the loop/cleanup interface, you don't need to do so.

    You may also use it as a context manager. Entering the context will start the thread, and
    exiting it will .terminate().join() it, in the following way:

    >>> a = MeGrimlock()
    >>> with a:
    >>>     ...
    >>> self.assertFalse(a.is_alive())
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._terminating = False

    def loop(self) -> None:
        """
        Run one iteration of the loop. Meant to be overrided. You do not need to override it
        if you decide to override run() through.

        This should block for as long as a single check will take, as termination checks take place
        between calls.
        """
        raise NotImplementedError('Override me!')

    def run(self) -> None:
        """
        Calls self.loop() indefinitely, until terminating condition is met
        """
        while not self._terminating:
            self.loop()
        self.cleanup()

    def cleanup(self):
        """
        Called after thread non-forced termination, in the thread's context.

        The default implementation does nothing.
        """

    def __enter__(self):
        """Returns self"""
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.terminate().join()
        return False

    def terminate(self, force: bool = False) -> 'TerminableThread':
        """
        Signal this thread to terminate.

        Forcing, if requested, will be done by injecting a SystemExit exception into target
        thread, so the thread must acquire GIL. For example, following would not be interruptable:

        >>> time.sleep(1000000)

        Note that calling force=True on PyPy won't work, and NotImplementedError will be raised
        instead.

        :param force: Whether to force a quit
        :return: self
        :raises RuntimeError: when something goes wrong with the underlying Python machinery
        :raises NotImplementedError: force=True was used on PyPy
        """
        self._terminating = True
        if force:
            if platform.python_implementation() == 'PyPy':
                raise NotImplementedError('force=True was made on PyPy')
            ret = ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self._ident),
                                                             ctypes.py_object(SystemExit))
            if ret == 0:
                ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self._ident), 0)
                raise RuntimeError('Zero threads killed!')
            elif ret > 1:
                ctypes.pythonapi.PyThreadState_SetAsyncExc(ctypes.c_long(self._ident), 0)
                raise RuntimeError('Multiple threads killed!')

        return self



from .choose import choose
from .iterators import infinite_counter, take_n, is_instance
from .sequences import is_last, add_next

__all__ = ['choose', 'infinite_counter', 'take_n', 'is_instance', 'is_last', 'add_next']



import functools
import typing as tp

__all__ = [
    'rethrow_as',
    'silence_excs',
    'catch_exception'
]


def silence_excs(*exc_types):
    """
    Silence given exception types.

    Can be either a decorator or a context manager
    """
    return rethrow_as(exc_types, None)


class rethrow_as:
    """
    Transform some exceptions into others.

    Either a decorator or a context manager

    New exception will be created by calling exception to transform to with
    repr of current one.

    You can also provide just two exceptions, eg.

    >>> rethrow_as(NameError, ValueError)

    You can also provide a pairwise translation, eg. from NameError to ValueError and from OSError
    to IOError

    >>> rethrow_as((NameError, ValueError), (OSError, IOError))

    If the second value is a None, exception will be silenced.
    """

    def __init__(self, *pairs,
                 exception_preprocessor: tp.Optional[tp.Callable[[Exception], str]] = None):
        """
        Pass tuples of (exception to catch - exception to transform to).

        :param exception_preprocessor: other callable/1 to use instead of repr.
            Should return a str, a text description of the exception
        """
        try:
            a, b = pairs  # throws ValueError
            op = issubclass(b, BaseException)  # throws TypeError
        except TypeError:
            op = b is None
        except ValueError:
            op = False

        if op:
            pairs = [pairs]

        self.mapping = list(pairs)
        self.exception_preprocessor = exception_preprocessor or repr

    def __call__(self, fun: tp.Callable) -> tp.Any:
        @functools.wraps(fun)
        def inner(*args, **kwargs):
            with self:
                return fun(*args, **kwargs)

        return inner

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            for from_, to in self.mapping:
                if issubclass(exc_type, from_):
                    if to is None:
                        return True
                    else:
                        raise to(self.exception_preprocessor(exc_val))


returning = tp.TypeVar('T')


def catch_exception(exc_class: tp.Union[tp.Type[Exception], tp.Tuple[tp.Type[Exception]]],
                    callable: tp.Callable[[], returning],
                    return_instead: tp.Optional[returning] = None,
                    return_value_on_no_exception: bool = False) -> tp.Union[Exception, returning]:
    """
    Catch exception of given type and return it. Functionally equivalent to:

    >>> try:
    >>>     callable()
    >>> except exc_class as e:
    >>>     result = e

    :param exc_class: Exception classes to catch
    :param callable: callable/0 to call to obtain it
    :param return_instead: what to return instead of the function result if it didn't end in an
        exception
    :param return_value_on_no_exception: whether to return the function result if exception didn't
        happen
    :raises ValueError: an exception was not thrown
    :raises TypeError: a different exception was thrown that the one we're catchin
    """
    try:
        result = callable()
    except exc_class as e:
        return e
    except Exception as e:
        raise TypeError('%s was thrown instead' % (e,))

    if return_instead is not None:
        return return_instead

    if return_value_on_no_exception:
        return result

    raise ValueError('Callable executed without error')



import logging
import typing as tp

logger = logging.getLogger(__name__)

__all__ = ['is_last']
T = tp.TypeVar('T')


# shamelessly copied from https://medium.com/better-programming/is-this-the-last-element-of-my-python-for-loop-784f5ff90bb5
def is_last(lst: tp.Iterable[T]) -> tp.Generator[tp.Tuple[bool, T], None, None]:
    """
    Return every element of the list, alongside a flag telling is this the last element.

    Use like:

    >>> for is_last, element in is_last(my_list):
    >>> if is_last:
    >>>     ...

    :param lst: list to iterate thru
    :return: a generator returning (bool, T)
    """
    iterable = iter(lst)
    ret_var = next(iterable)
    for val in iterable:
        yield False, ret_var
        ret_var = val
    yield True, ret_var


def add_next(lst: tp.Iterable[T]) -> tp.Generator[tp.Tuple[T, tp.Optional[T]], None, None]:
    """
    Yields a 2-tuple of given iterable, presenting the next element as second element of the tuple.

    The last element will be the last element alongside with a None

    Example:

    >>> list(add_next([1, 2, 3, 4, 5])) == [(1, 2), (2, 3), (3, 4), (4, 5), (5, None)]
    """
    iterator = iter(lst)
    try:
        prev_val = next(iterator)
    except StopIteration:
        return
    for val in iterator:
        yield prev_val, val
        prev_val = val
    yield prev_val, None




import typing as tp

from satella.configuration.schema import Descriptor, descriptor_from_dict
from satella.exceptions import ConfigurationValidationError

__all__ = ['DictObject', 'apply_dict_object']


class DictObject(dict):
    """
    A dictionary wrapper that can be accessed by attributes. Eg:

    >>> a = DictObject({'test': 5})
    >>> self.assertEqual(a.test, 5)
    """

    def __getattr__(self, item: str) -> tp.Any:
        try:
            return self[item]
        except KeyError as e:
            raise AttributeError(repr(e))

    def __setattr__(self, key: str, value: tp.Any) -> None:
        self[key] = value

    def __delattr__(self, key: str) -> None:
        try:
            del self[key]
        except KeyError as e:
            raise AttributeError(repr(e))

    def is_valid_schema(self, schema: tp.Optional[tp.Union[Descriptor, dict]] = None,
                        **kwarg_schema) -> bool:
        """
        Check if this dictionary conforms to particular schema.

        Schema is either a Descriptor, or a JSON-based schema. See satella.configuration.schema for details.
        Schema can be passed as well using kwargs. It will be preferred to the one passed as schema.

        :param schema: schema to verify against
        :return: whether is conformant
        """

        if kwarg_schema:
            schema = kwarg_schema

        if isinstance(schema, Descriptor):
            descriptor = schema
        else:
            descriptor = descriptor_from_dict(schema)

        try:
            descriptor(self)
        except ConfigurationValidationError:
            return False
        else:
            return True


def apply_dict_object(v: tp.Any) -> tp.Union[DictObject, tp.Any]:
    """
    Apply DictObject() to every dict inside v.

    This assumes that the only things that will be touched will be nested dicts and lists.

    If you pass a non-dict and a non-list, they will be returned as is.
    """
    if isinstance(v, list):
        return [apply_dict_object(x) for x in v]
    elif isinstance(v, dict):
        return DictObject({
            k: apply_dict_object(val) for k, val in v.items()
        })
    else:
        return v



"""
Handle signals, like a boss
"""
import signal
import time

import typing as tp

end = False


def __sighandler(a, b):
    global end
    end = True


def hang_until_sig(extra_signals: tp.Optional[tp.List[int]] = None,
                   end_on_keyboard_interrupt: bool = True):
    """
    Will hang until this process receives SIGTERM or SIGINT.
    If you pass extra signal IDs (signal.SIG*) with extra_signals,
    then also on those signals this call will release.

    :param extra_signals: a list of extra signals to listen to
    :param end_on_keyboard_interrupt: whether to consider receiving a KeyboardInterrupt as
        a signal to finish
    """
    extra_signals = extra_signals or []
    global end

    signal.signal(signal.SIGTERM, __sighandler)
    signal.signal(signal.SIGINT, __sighandler)
    for s in extra_signals:
        signal.signal(s, __sighandler)

    while not end:
        try:
            time.sleep(0.5)
        except KeyboardInterrupt:
            if end_on_keyboard_interrupt:
                end = True

    end = False  # reset for next use
